/*
 * Copyright (C) 1997-2007 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file command.c does client authentication and executes server commands */

#include <sys/time.h> /* gettimeofday */
#include "server.cmdline.h"
#include "afs.h"
#include "server.h"
#include "vss.h"
#include "send.h"
#include "rc4.h"
#include <openssl/rc4.h>
#include "error.h"
#include "net.h"
#include "daemon.h"
#include "string.h"
#include "fd.h"
#include "user_list.h"
#include "server_command_list.h"

/** commands including options must be shorter than this */
#define MAX_COMMAND_LEN 4096

static RC4_KEY rc4_recv_key;
static RC4_KEY rc4_send_key;
static unsigned char rc4_buf[2 * RC4_KEY_LEN];

extern const char *status_item_list[NUM_STAT_ITEMS];
extern struct misc_meta_data *mmd;
extern struct audio_file_selector selectors[];
extern struct sender senders[];

static void dummy(__a_unused int s)
{}

static void mmd_dup(struct misc_meta_data *new_mmd)
{
	mmd_lock();
	*new_mmd = *mmd;
	mmd_unlock();
}

/*
 * compute human readable string containing
 * vss status for given integer value
 */
static char *vss_status_tohuman(unsigned int flags)
{
	if (flags & VSS_PLAYING)
		return para_strdup("playing");
	else if (flags & VSS_NEXT)
		return para_strdup("stopped");
	else
		return para_strdup("paused");
}

/*
 * return human readable permission string. Never returns NULL.
 */
static char *cmd_perms_itohuman(unsigned int perms)
{
	char *msg = para_malloc(5 * sizeof(char));

	msg[0] = perms & AFS_READ? 'a' : '-';
	msg[1] = perms & AFS_WRITE? 'A' : '-';
	msg[2] = perms & VSS_READ? 'v' : '-';
	msg[3] = perms & VSS_WRITE? 'V' : '-';
	msg[4] = '\0';
	return msg;
}

/*
 * Never returns NULL.
 */
static char *vss_get_status_flags(unsigned int flags)
{
	char *msg = para_malloc(5 * sizeof(char));

	msg[0] = (flags & VSS_PLAYING)? 'P' : '_';
	msg[1] = (flags & VSS_NOMORE)? 'O' : '_';
	msg[2] = (flags & VSS_NEXT)? 'N' : '_';
	msg[3] = (flags & VSS_REPOS)? 'R' : '_';
	msg[4] = '\0';
	return msg;
}

/*
 * compute status bar string. Never returns NULL
 */
static char *get_sb_string(struct misc_meta_data *nmmd)
{
	char *base, *ret;
	long long unsigned secs = 0, rsecs = 0, percent = 0;

	base = para_basename(nmmd->filename);
	if (!base)
		return para_strdup("");
	if (!base[0])
		return base;
	if (nmmd->afi.chunks_total) {
		secs = (long long) nmmd->afi.seconds_total * nmmd->chunks_sent
			/ nmmd->afi.chunks_total;
		rsecs = (long long) nmmd->afi.seconds_total *
			(nmmd->afi.chunks_total - nmmd->chunks_sent)
			/ nmmd->afi.chunks_total;
		percent = 100 * ((nmmd->chunks_sent + 5) / 10)
			/ ((nmmd->afi.chunks_total + 5) / 10);
	}
	ret = make_message("%llu:%02llu [%llu:%02llu] (%llu%%) %s",
		secs / 60, secs % 60,
		rsecs / 60, rsecs % 60,
		percent,
		base
	);
	free(base);
	return ret;
}

static char *get_status(struct misc_meta_data *nmmd)
{
	char *bar, *ret, mtime[30] = "";
	char *status, *flags; /* vss status info */
	char *ut = uptime_str();
	long offset = (nmmd->offset + 500) / 1000;
	struct timeval now;
	struct tm mtime_tm;

	if (nmmd->audio_format >= 0) {
		localtime_r(&nmmd->mtime, &mtime_tm);
		strftime(mtime, 29, "%a %b %d %Y", &mtime_tm);
	}
	/* report real status */
	status = vss_status_tohuman(nmmd->vss_status_flags);
	flags = vss_get_status_flags(nmmd->vss_status_flags);
	bar = para_basename(nmmd->filename);
	gettimeofday(&now, NULL);
	ret = make_message(
		"%s:%lu\n"	"%s:%s\n"		"%s:%i\n"	"%s:%u\n"
		"%s:%s\n"	"%s:%s\n"	"%s:%s\n"	"%s:%s\n"
		"%s:%li\n" 	"%s:%s\n"	"%s"		"%s"
		"%s:%s\n"	"%s:%lu.%lu\n"	"%s:%lu.%lu\n",
		status_item_list[SI_FILE_SIZE], nmmd->size / 1024,
		status_item_list[SI_MTIME], mtime,
		status_item_list[SI_LENGTH], nmmd->afi.seconds_total,
		status_item_list[SI_NUM_PLAYED], nmmd->num_played,

		status_item_list[SI_STATUS_BAR], bar ? bar : "(none)",
		status_item_list[SI_STATUS], status,
		status_item_list[SI_STATUS_FLAGS], flags,
		status_item_list[SI_SELECTOR], selectors[nmmd->selector_num].name,

		status_item_list[SI_OFFSET], offset,
		status_item_list[SI_FORMAT], audio_format_name(nmmd->audio_format),
		nmmd->selector_info,
		nmmd->afi.info_string,

		status_item_list[SI_UPTIME], ut,
		status_item_list[SI_STREAM_START],
			(long unsigned)nmmd->stream_start.tv_sec,
			(long unsigned)nmmd->stream_start.tv_usec,
		status_item_list[SI_CURRENT_TIME],
			(long unsigned)now.tv_sec,
			(long unsigned)now.tv_usec

	);
	free(bar);
	free(flags);
	free(status);
	free(ut);
	return ret;
}

static int check_sender_args(int argc, char **argv, struct sender_command_data *scd)
{
	int i;
	/* this has to match sender.h */
	const char *subcmds[] = {"add", "delete", "allow", "deny", "on", "off", NULL};

	scd->sender_num = -1;
	if (argc < 2)
		return -E_COMMAND_SYNTAX;
	for (i = 0; senders[i].name; i++)
		if (!strcmp(senders[i].name, argv[1]))
			break;
	PARA_DEBUG_LOG("%d:%s\n", argc, argv[1]);
	if (!senders[i].name)
		return -E_COMMAND_SYNTAX;
	scd->sender_num = i;
	for (i = 0; subcmds[i]; i++)
		if (!strcmp(subcmds[i], argv[2]))
			break;
	if (!subcmds[i])
		return -E_COMMAND_SYNTAX;
	scd->cmd_num = i;
	mmd_lock();
	if (!senders[scd->sender_num].client_cmds[scd->cmd_num]) {
		mmd_unlock();
		return -E_SENDER_CMD;
	}
	mmd_unlock();
	switch (scd->cmd_num) {
	case SENDER_ON:
	case SENDER_OFF:
		if (argc != 3)
			return -E_COMMAND_SYNTAX;
		break;
	case SENDER_DENY:
	case SENDER_ALLOW:
		if (argc != 4 && argc != 5)
			return -E_COMMAND_SYNTAX;
		if (!inet_aton(argv[3], &scd->addr))
			return -E_COMMAND_SYNTAX;
		scd->netmask = 32;
		if (argc == 5) {
			scd->netmask = atoi(argv[4]);
			if (scd->netmask < 0 || scd->netmask > 32)
				return -E_COMMAND_SYNTAX;
		}
		break;
	case SENDER_ADD:
	case SENDER_DELETE:
		if (argc != 4 && argc != 5)
			return -E_COMMAND_SYNTAX;
		if (!inet_aton(argv[3], &scd->addr))
			return -E_COMMAND_SYNTAX;
		scd->port = -1;
		if (argc == 5) {
			scd->port = atoi(argv[4]);
			if (scd->port < 0 || scd->port > 65535)
				return -E_COMMAND_SYNTAX;
		}
		break;
	default:
		return -E_COMMAND_SYNTAX;
	}
	return 1;
}

int com_sender(int fd, int argc, char **argv)
{
	int i, ret;
	struct sender_command_data scd;

	if (argc < 2) {
		char *msg = NULL;
		for (i = 0; senders[i].name; i++) {
			char *tmp = make_message("%s%s\n",
				msg? msg : "", senders[i].name);
			free(msg);
			msg = tmp;
		}
		ret = send_buffer(fd, msg);
		free(msg);
		return ret;
	}
	ret = check_sender_args(argc, argv, &scd);
	if (ret < 0) {
		char *msg;
		if (scd.sender_num < 0)
			return ret;
		msg = senders[scd.sender_num].help();
		send_buffer(fd, msg);
		free(msg);
		return 1;
	}
	for (i = 0; i < 10; i++) {
		mmd_lock();
		if (mmd->sender_cmd_data.cmd_num >= 0) {
			mmd_unlock();
			usleep(100 * 1000);
			continue;
		}
		mmd->sender_cmd_data = scd;
		mmd_unlock();
		break;
	}
	return (i < 10)? 1 : -E_LOCK;
}

/* server info */
int com_si(int fd, int argc, __a_unused char **argv)
{
	int i, ret;
	char *ut;
	char *selector_string = NULL, *sender_info = NULL, *sender_list = NULL;

	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mmd_lock();
	for (i = 0; selectors[i].name; i++) {
		selector_string = para_strcat(selector_string, selectors[i].name);
		selector_string = para_strcat(selector_string, " ");
	}
	for (i = 0; senders[i].name; i++) {
		char *info = senders[i].info();
		sender_info = para_strcat(sender_info, info);
		free(info);
		sender_list = para_strcat(sender_list, senders[i].name);
		sender_list = para_strcat(sender_list, " ");
	}
	ut = uptime_str();
	ret = send_va_buffer(fd, "up: %s\nplayed: %u\n"
		"pid: %d\n"
		"connections (active/accepted/total): %u/%u/%u\n"
		"current loglevel: %i\n"
		"supported audio file selectors: %s\n"
		"supported audio formats: %s\n"
		"supported senders: %s\n"
		"%s",
		ut, mmd->num_played,
		getppid(),
		mmd->active_connections,
		mmd->num_commands,
		mmd->num_connects,
		conf.loglevel_arg,
		selector_string,
		supported_audio_formats(),
		sender_list,
		sender_info
	);
	mmd_unlock();
	free(ut);
	free(selector_string);
	free(sender_list);
	free(sender_info);
	return ret;
}

/* version */
int com_version(int fd, int argc, __a_unused char **argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	return send_buffer(fd, VERSION_TEXT("server")
		"built: " BUILD_DATE "\n"
		SYSTEM ", " CC_VERSION "\n"
	);
}

/* sc */
int com_sc(int fd, int argc, char **argv)
{
	char *name = NULL;
	int ret, old = 0, count = -1; /* print af change forever */

	if (argc > 1)
		count = atoi(argv[1]);
repeat:
	mmd_lock();
	if (old != mmd->num_played) {
		old = mmd->num_played;
		name = para_strdup(mmd->filename);
	}
	mmd_unlock();
	if (name) {
		ret = send_va_buffer(fd, "%s\n", name);
		free(name);
		name = NULL;
		if (ret < 0)
			return ret;
		if (argc > 1 && !--count)
			return 1;
	}
	usleep(500000);
	goto repeat;
}

/* sb */
int com_sb(int fd, int argc, char **argv)
{
	char *sb;
	int ret, nr = -1;	/* status bar will be printed that many
				 * times. Negative value means: print
				 * forever
				 */
	if (argc > 1)
		nr = atoi(argv[1]);
	while (nr) {
		mmd_lock();
		sb = get_sb_string(mmd);
		mmd_unlock();
		ret = send_va_buffer(fd, "%s\n", sb);
		free(sb);
		if (ret < 0)
			return ret;
		if (nr == 1)
			return 1;
		usleep(500000);
		if (nr > 0)
			nr--;
	}
	return 1;
}

/* stat */
int com_stat(int fd, int argc, char **argv)
{
	int ret, num = 0;/* status will be printed that many
			  * times. num <= 0 means: print forever
			  */
	struct misc_meta_data tmp, *nmmd = &tmp;
	char *s;

	signal(SIGUSR1, dummy);

	if (argc > 1)
		num = atoi(argv[1]);
	for (;;) {

		mmd_dup(nmmd);
		s = get_status(nmmd);
		ret = send_buffer(fd, s);
		free(s);
		if (ret < 0)
			goto out;
		ret = 1;
		if (num == 1)
			goto out;
		sleep(50);
		if (getppid() == 1)
			return -E_SERVER_CRASH;
	}
out:
	return ret;
}

static int send_list_of_commands(int fd, struct server_command *cmd,
		const char *handler)
{
	int ret, i;

	for (i = 1; cmd->name; cmd++, i++) {
		char *perms = cmd_perms_itohuman(cmd->perms);
		ret = send_va_buffer(fd, "%s\t%s\t%s\t%s\n", cmd->name,
			handler,
			perms,
			cmd->description);
		free(perms);
		if (ret < 0)
			return ret;
	}
	return 1;
}

/* always returns string that must be freed by the caller in handler */
static struct server_command *get_cmd_ptr(char *name, char **handler)
{
	struct server_command *cmd;

	for (cmd = server_cmds; cmd->name; cmd++)
		if (!strcmp(cmd->name, name)) {
			if (handler)
				*handler = para_strdup("para_server"); /* server commands */
			return cmd;
		}
	/* not found, look for commands supported by the current selector */
	mmd_lock();
	if (handler)
		*handler = make_message("the %s selector",
			selectors[mmd->selector_num].name);
	cmd = selectors[mmd->selector_num].cmd_list;
	mmd_unlock();
	for (; cmd->name; cmd++)
		if (!strcmp(cmd->name, name))
			return cmd;
	return NULL;
}

/* help */
int com_help(int fd, int argc, char **argv)
{
	struct server_command *cmd;
	char *perms, *handler;
	int ret;

	if (argc < 2) {
		/* no argument given, print list of commands */
		if ((ret = send_list_of_commands(fd, server_cmds, "server")) < 0)
			return ret;
		mmd_lock();
		handler = para_strdup(selectors[mmd->selector_num].name);
		cmd = selectors[mmd->selector_num].cmd_list;
		mmd_unlock();
		ret = send_list_of_commands(fd, cmd, handler);
		free(handler);
		return ret;
	}
	/* argument given for help */
	cmd = get_cmd_ptr(argv[1], &handler);
	if (!cmd) {
		free(handler);
		return -E_BAD_CMD;
	}
	perms = cmd_perms_itohuman(cmd->perms);
	ret = send_va_buffer(fd,
		"%s - %s\n\n"
		"handler: %s\n"
		"permissions: %s\n"
		"usage: %s\n\n"
		"%s\n",
		argv[1],
		cmd->description,
		handler,
		perms,
		cmd->usage,
		cmd->help
	);
	free(perms);
	free(handler);
	return ret;
}

/* hup */
int com_hup(__a_unused int fd, int argc, __a_unused char **argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	kill(getppid(), SIGHUP);
	return 1;
}

/* term */
int com_term(__a_unused int fd, int argc, __a_unused char **argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	kill(getppid(), SIGTERM);
	return 1;
}

int com_play(__a_unused int fd, int argc, __a_unused char **argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mmd_lock();
	mmd->new_vss_status_flags |= VSS_PLAYING;
	mmd->new_vss_status_flags &= ~VSS_NOMORE;
	mmd_unlock();
	return 1;

}

/* stop */
int com_stop(__a_unused int fd, int argc, __a_unused char **argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mmd_lock();
	mmd->new_vss_status_flags &= ~VSS_PLAYING;
	mmd->new_vss_status_flags &= ~VSS_REPOS;
	mmd->new_vss_status_flags |= VSS_NEXT;
	mmd_unlock();
	return 1;
}

/* pause */
int com_pause(__a_unused int fd, int argc, __a_unused char **argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mmd_lock();
	if (!vss_paused())
		mmd->events++;
	mmd->new_vss_status_flags &= ~VSS_PLAYING;
	mmd->new_vss_status_flags &= ~VSS_NEXT;
	mmd_unlock();
	return 1;
}

int com_chs(int fd, int argc, char **argv)
{
	int i, ret;

	if (argc == 1) {
		char *selector;
		mmd_lock();
		selector = para_strdup(selectors[mmd->selector_num].name);
		mmd_unlock();
		ret = send_va_buffer(fd, "%s\n", selector);
		free(selector);
		return ret;
	}
	for (i = 0; selectors[i].name; i++) {
		if (strcmp(selectors[i].name, argv[1]))
			continue;
		mmd_lock();
		mmd->selector_change = i;
		mmd->events++;
		mmd_unlock();
		return 1;
	}
	return -E_BAD_SELECTOR;
}

/* next */
int com_next(__a_unused int fd, int argc, __a_unused char **argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mmd_lock();
	mmd->events++;
	mmd->new_vss_status_flags |= VSS_NEXT;
	mmd_unlock();
	return 1;
}

/* nomore */
int com_nomore(__a_unused int fd, int argc, __a_unused char **argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mmd_lock();
	if (vss_playing() || vss_paused())
		mmd->new_vss_status_flags |= VSS_NOMORE;
	mmd_unlock();
	return 1;
}

/* ff */
int com_ff(__a_unused int fd, int argc, char **argv)
{
	long promille;
	int ret, backwards = 0;
	unsigned i;
	char c;

	if (argc != 2)
		return -E_COMMAND_SYNTAX;
	if (!(ret = sscanf(argv[1], "%u%c", &i, &c)))
		return -E_COMMAND_SYNTAX;
	if (ret > 1 && c == '-')
		backwards = 1; /* jmp backwards */
	mmd_lock();
	ret = -E_NO_AUDIO_FILE;
	if (!mmd->afi.chunks_total || !mmd->afi.seconds_total)
		goto out;
	promille = (1000 * mmd->current_chunk) / mmd->afi.chunks_total;
	if (backwards)
		promille -= 1000 * i / mmd->afi.seconds_total;
	else
		promille += 1000 * i / mmd->afi.seconds_total;
	if (promille < 0)
		promille = 0;
	if (promille >  1000) {
		mmd->new_vss_status_flags |= VSS_NEXT;
		goto out;
	}
	mmd->repos_request = (mmd->afi.chunks_total * promille) / 1000;
	mmd->new_vss_status_flags |= VSS_REPOS;
	mmd->new_vss_status_flags &= ~VSS_NEXT;
	mmd->events++;
	ret = 1;
out:
	mmd_unlock();
	return ret;
}

/* jmp */
int com_jmp(__a_unused int fd, int argc, char **argv)
{
	long unsigned int i;
	int ret;

	if (argc != 2)
		return -E_COMMAND_SYNTAX;
	if (sscanf(argv[1], "%lu", &i) <= 0)
		return -E_COMMAND_SYNTAX;
	mmd_lock();
	ret = -E_NO_AUDIO_FILE;
	if (!mmd->afi.chunks_total)
		goto out;
	if (i > 100)
		i = 100;
	PARA_INFO_LOG("jumping to %lu%%\n", i);
	mmd->repos_request = (mmd->afi.chunks_total * i + 50)/ 100;
	PARA_INFO_LOG("sent: %lu,  offset before jmp: %lu\n",
		mmd->chunks_sent, mmd->offset);
	mmd->new_vss_status_flags |= VSS_REPOS;
	mmd->new_vss_status_flags &= ~VSS_NEXT;
	ret = 1;
	mmd->events++;
out:
	mmd_unlock();
	return ret;
}

/*
 * check if perms are sufficient to exec a command having perms cmd_perms.
 * Returns 0 if perms are sufficient, -E_PERM otherwise.
 */
static int check_perms(unsigned int perms, struct server_command *cmd_ptr)
{
	PARA_DEBUG_LOG("%s", "checking permissions\n");
	return (cmd_ptr->perms & perms) < cmd_ptr->perms ? -E_PERM : 0;
}

/*
 * Parse first string from *cmd and lookup in table of valid commands.
 * On error, NULL is returned.
 */
static struct server_command *parse_cmd(const char *cmdstr)
{
	char buf[255];
	int n = 0;

	sscanf(cmdstr, "%200s%n", buf, &n);
	if (!n)
		return NULL;
	buf[n] = '\0';
	return get_cmd_ptr(buf, NULL);
}

static long int para_rand(long unsigned max)
{
	return (long int) ((max + 0.0) * (random() / (RAND_MAX + 1.0)));
}

static void init_rc4_keys(void)
{
	int i;

	for (i = 0; i < 2 * RC4_KEY_LEN; i++)
		rc4_buf[i] = para_rand(256);
	PARA_DEBUG_LOG("rc4 keys initialized (%u:%u)\n",
		(unsigned char) rc4_buf[0],
		(unsigned char) rc4_buf[RC4_KEY_LEN]);
	RC4_set_key(&rc4_recv_key, RC4_KEY_LEN, rc4_buf);
	RC4_set_key(&rc4_send_key, RC4_KEY_LEN, rc4_buf + RC4_KEY_LEN);
}

static void rc4_recv(unsigned long len, const unsigned char *indata,
		unsigned char *outdata, __a_unused void *private_data)
{
	RC4(&rc4_recv_key, len, indata, outdata);
}

static void rc4_send(unsigned long len, const unsigned char *indata,
		unsigned char *outdata, __a_unused void *private_data)
{
	RC4(&rc4_send_key, len, indata, outdata);
}

/**
 * perform user authentication and execute a command
 *
 * \param fd the file descriptor to send output to
 * \param addr socket address info of peer
 *
 * \return EXIT_SUCCESS or EXIT_FAILURE
 *
 * Whenever para_server accepts an incoming tcp connection on
 * the port it listens on, it forks and the resulting child
 * calls this function.
 *
 * An RSA-based challenge/response is used to authenticate
 * the peer. It that authentication succeeds, a random RC4
 * session key is generated and sent back to the peer,
 * encrypted with its RSA public key.  From this point on,
 * all transfers are crypted with this session key.
 *
 * Next it is checked if the peer supplied  a valid server
 * command or a command for the audio file selector currently
 * in use.  If yes, and if the user has sufficient
 * permissions to execute that command, the function calls
 * the corresponding command handler which does argument
 * checking and further processing.
 *
 * In order to cope with a DOS attacks, a timeout is set up
 * which terminates the function if the connection was not
 * authenticated when the timeout expires.
 *
 * \sa alarm(2), rc4(3), crypt.c, crypt.h
 */
int handle_connect(int fd, struct sockaddr_in *addr)
{
	int numbytes, ret, argc, use_rc4 = 0;
	char buf[4096];
	unsigned char crypt_buf[MAXLINE];
	struct user *u;
	struct server_command *cmd = NULL;
	long unsigned challenge_nr, chall_response;
	char **argv = NULL;
	char *p, *command = NULL;

	signal(SIGCHLD, SIG_IGN);
	signal(SIGINT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);
	signal(SIGHUP, SIG_DFL);
	signal(SIGUSR1, SIG_IGN);

	challenge_nr = random();
	/* send Welcome message */
	ret = send_va_buffer(fd, "This is para_server, version "
		PACKAGE_VERSION  ".\n" );
	if (ret < 0)
		goto err_out;
	/* recv auth request line */
	ret = recv_buffer(fd, buf, sizeof(buf));
	if (ret < 0)
		goto err_out;
	if (ret <= 6) {
		ret = -E_AUTH;
		goto err_out;
	}
	numbytes = ret;
	ret = -E_AUTH;
	if (strncmp(buf, "auth ", 5))
		goto err_out;

	if (numbytes < 9 || strncmp(buf, "auth rc4 ", 9))
		p = buf + 5; /* client version < 0.2.6 */
	else {
		p = buf + 9; /* client version >= 0.2.6 */
		use_rc4 = 1;
	}
	PARA_DEBUG_LOG("received %s request for user %s\n",
		use_rc4? "rc4" : "auth", p);
	ret = -E_BAD_USER;
	u = lookup_user(p);
	if (!u)
		goto err_out;
	ret = para_encrypt_challenge(u->rsa, challenge_nr, crypt_buf);
	if (ret <= 0)
		goto err_out;
	numbytes = ret;
	PARA_DEBUG_LOG("sending %d byte challenge\n", numbytes);
	/* We can't use send_buffer here since buf may contain null bytes */
	ret = send_bin_buffer(fd,(char *) crypt_buf, numbytes);
	if (ret < 0)
		goto err_out;
	/* recv decrypted number */
	numbytes = recv_buffer(fd, buf, sizeof(buf));
	ret = numbytes;
	if (ret < 0)
		goto err_out;
	ret = -E_AUTH;
	if (!numbytes)
		goto err_out;
	if (sscanf(buf, CHALLENGE_RESPONSE_MSG "%lu", &chall_response) < 1
			|| chall_response != challenge_nr)
		goto err_out;
	/* auth successful. Send 'Proceed' message */
	PARA_INFO_LOG("good auth for %s (%lu)\n", u->name, challenge_nr);
	sprintf(buf, "%s", PROCEED_MSG);
	if (use_rc4) {
		init_rc4_keys();
		ret = para_encrypt_buffer(u->rsa, rc4_buf, 2 * RC4_KEY_LEN,
			(unsigned char *)buf + PROCEED_MSG_LEN + 1);
		if (ret <= 0)
			goto err_out;
		numbytes = ret + strlen(PROCEED_MSG) + 1;
	} else
		numbytes = strlen(buf);
	ret = send_bin_buffer(fd, buf, numbytes);
	if (ret < 0)
		goto err_out;
	if (use_rc4)
		enable_crypt(fd, rc4_recv, rc4_send, NULL);
	/* read command */
	while ((numbytes = recv_buffer(fd, buf, sizeof(buf))) > 0) {
//		PARA_INFO_LOG("recvd: %s (%d)\n", buf, numbytes);
		ret = -E_COMMAND_SYNTAX;
		if (command && numbytes + strlen(command) > MAX_COMMAND_LEN) /* DOS */
			goto err_out;
		command = para_strcat(command, buf);
		if ((p = strstr(command, EOC_MSG))) {
			*p = '\0';
			break;
		}
	}
	ret = numbytes;
	if (ret < 0)
		goto err_out;
	ret = -E_BAD_CMD;
	/* parse command */
	if (!(cmd = parse_cmd(command)))
		goto err_out;
	/* valid command, check permissions */
	ret = check_perms(u->perms, cmd);
	if (ret < 0)
		goto err_out;
	/* valid command and sufficient perms */
	alarm(0);
	argc = split_args(command, &argv, "\n");
	mmd_lock();
	mmd->num_commands++;
	mmd_unlock();
	PARA_NOTICE_LOG("calling com_%s() for %s@%s\n", cmd->name, u->name,
		inet_ntoa(addr->sin_addr));
	ret = cmd->handler(fd, argc, argv);
	if (ret >= 0) {
		ret = EXIT_SUCCESS;
		goto out;
	}
err_out:
	if (ret != -E_SEND && ret != -E_RECV) {
		PARA_NOTICE_LOG("%s\n", PARA_STRERROR(-ret));
		send_va_buffer(fd, "%s\n", PARA_STRERROR(-ret));
	}
	ret = EXIT_FAILURE;
out:
	free(command);
	free(argv);
	mmd_lock();
	if (cmd && (cmd->perms & AFS_WRITE) && ret >= 0)
		mmd->events++;
	mmd->active_connections--;
	mmd_unlock();
	return ret;
}
